/*
    VDP1 Framebuffer Viewing Utility
    by Charles MacDonald
    WWW: http://cgfm2.emuviews.com

    Default Comms Link port is $320, program uses what's in the "AR_PORT"
    environment variable if present.
*/

#include "main.h"

uint16 comms_port = 0x320;
uint8 frame[0x40000];
uint8 cram[0x1000];
uint8 name[128];

uint16 palette[0x800];
BITMAP *bmp, *tmp;
PALETTE pal;

int main (int argc, char *argv[])
{
    int clip = 0;
    int width, height;
    int i, x, y;
    uint8 *p;

    /* Check if the AR_PORT environment variable exists */
    p = getenv("AR_PORT");
    if(p)
        comms_port = strtol(p, NULL, 16);

    /* Read VDP1 framebuffer and color RAM */
    printf("Reading VDP memory, please wait...\n");
    comms_read_memory(frame, 0x25C80000, 0x40000);
    comms_read_memory(cram, 0x25F00000, 0x1000);
    comms_read_memory(name, 0x26002060, 0x30);

    /* Make sure data port is cleared */
    outp(comms_port, 0x00);

    /* Initialize Allegro */
    allegro_init();
    set_color_depth(16);

    bswap(cram, 0x1000);
    for(i = 0; i < 2048; i++)
    {
        uint16 c = *(uint16 *)&cram[i << 1];
        int r = (c >>  0) & 0x1F;
        int g = (c >>  5) & 0x1F;
        int b = (c >> 10) & 0x1F;
        palette[i] = makecol16(r<<3,g<<3,b<<3);
    }

    /* Set up video display */
    bmp = create_bitmap(512, 256);

    /* Write out framebuffer data */
    bswap(frame, 0x40000);
    for(y = 0; y < 256; y++)
    for(x = 0; x < 512; x++)
    {
        uint16 d = *(uint16 *)&frame[(y*1024)+(x*2)];

        if(d & 0x8000)
        {
                int r = (d >>  0) & 0x1F;
                int g = (d >>  5) & 0x1F;
                int b = (d >> 10) & 0x1F;
                putpixel(bmp, x, y, makecol16(r<<3,g<<3,b<<3));
        }
        else
        {
                putpixel(bmp, x, y, palette[d & 0x07FF]);
        }

        /* Fill VDP2 graphics area with bright pink */
        if(d == 0x0000)
        {
            putpixel(bmp, x, y, makecol16(0xFF, 0x00, 0xFF));
        }

        /* Check for unused memory */
        if((d == 0xAAAA) || (d == 0x5555))
        {
            if((x >= 320) && (x < 352) && (y < 224))
                clip |= 1;
            if((y >= 224) && (y < 240) && (x < 320))
                clip |= 2;
        } 
    }

    /* Determine size of display area */
    width = (clip & 1) ? 320 : 352;
    height = (clip & 2) ? 224 : 240;

    /* Make PCX snapshot */
    tmp = create_bitmap(width, height);
    blit(bmp, tmp, 0, 0, 0, 0, width, height);
    get_palette(pal);

    /* Use game title as filename */
    i = 0x2F;
    while(name[i] == ' ')
    {
        name[i] = '\0';
        i--;
    }
    printf("[%s]\n", name);
    strcat(name, ".pcx");
    save_bitmap(name, tmp, pal);

    /* Shut down */
    destroy_bitmap(tmp);
    destroy_bitmap(bmp);

    return 0;
}

int comms_read_memory(uint8 *buffer, uint32 address, uint32 length)
{
    int i, error = 1;
    uint8 checksum[2];

    /* Synchronize with Saturn */
    while (1)
    {
        if (comms_exchange_byte('I') != 'D')
            continue;
        if (comms_exchange_byte('N') == 'O')
            break;
    }

    /* Function $01 - Download data */
    comms_exchange_byte(1);

    /* Ignore R9 value (status?) */
    comms_receive_longword();

    /* Send address and size of memory to get */
    comms_send_longword(address);
    comms_send_longword(length);

    /* Receive data from Saturn */
    for (i = 0; i < length; i++)
    {
        buffer[i] = comms_exchange_byte(0);
        checksum[0] += buffer[i];
    }

    /* Compare checksums */
    checksum[1] = comms_exchange_byte(0);
    if(checksum[0] != checksum[1])
    {
        printf("Error: checksum mismatch (PC:%02X SAT:%04X)\n", checksum[0], checksum[1]);
        error = 0;
    }

    /* Send dummy address and zero length to end transfer */
    comms_send_longword(0);
    comms_send_longword(0);

    /* Read 'OK' characters */
    if(comms_exchange_byte(0) != 'O') error = 0;
    if(comms_exchange_byte(0) != 'K') error = 0;

    return error;
}


uint8 comms_exchange_byte(uint8 data)
{
    outp(comms_port, data);
    while (inp(comms_port + 2) & 1);
    return inp(comms_port);
}

void comms_send_longword(uint32 data)
{
    comms_exchange_byte(data >> 24);
    comms_exchange_byte((data >> 16) & 0xff);
    comms_exchange_byte((data >> 8) & 0xff);
    comms_exchange_byte(data & 0xff);
}

uint32 comms_receive_longword(void)
{
    uint32  data;
    data = comms_exchange_byte(0) << 24;
    data |= comms_exchange_byte(0) << 16;
    data |= comms_exchange_byte(0) << 8;
    data |= comms_exchange_byte(0);
    return data;
}

void bswap(uint8 *mem, int length)
{
    int i;
    for(i = 0; i < length; i += 2)
    {
        uint8 temp = mem[i+0];
        mem[i+0] = mem[i+1];
        mem[i+1] = temp;
    }
}

